from django.contrib import admin

from .models import Profile


class ProfileAdmin(admin.ModelAdmin):
    search_fields = ('id', 'sso_id', 'user__username', 'user__email')
    list_display = ('user', 'sso_id', 'sso_rev', 'created_at', 'created_by_event')
    filter_horizontal = ('groups', )


admin.site.register(Profile, ProfileAdmin)



import logging
import json
import responses

from django.contrib.auth import get_user_model
from django.core import mail
from django.urls import reverse

from rest_framework import status

from django_sso_app.core.tests.factories import UserTestCase

from .... import app_settings
from ...emails.models import EmailAddress
from ...groups.models import Group
from ...profiles.models import Profile
from ..models import RequestForCredentialsEventListener
from ..tasks import create_users_by_events

logger = logging.getLogger(__name__)
User = get_user_model()
ELASTICSEARCH_INDEX_URL = app_settings.EVENTS_HOST + '/{}/_search'.format(app_settings.EVENTS_INDEX)


class EventTestCase(UserTestCase):
    @responses.activate
    def test_can_create_users_by_event_objects(self):
        email_verified = True
        default_password = 'password1234'
        rfc_event_listener = RequestForCredentialsEventListener(confirm_email=email_verified,
                                                                first_name_field='name',
                                                                default_password=default_password)
        rfc_event_listener.save()

        new_user_email = 'pinco@example.com'
        new_user_first_name = 'Pinco'
        event_group_name = 'q-hack40-cps-bs'
        event_password = 'qhack'

        _es_response = """{
          "took" : 1,
          "timed_out" : false,
          "_shards" : {
            "total" : 1,
            "successful" : 1,
            "skipped" : 0,
            "failed" : 0
          },
          "hits" : {
            "total" : {
              "value" : 1,
              "relation" : "eq"
            },
            "max_score" : null,
            "hits" : [
              {
                "_index" : "events_private",
                "_type" : "_doc",
                "_id" : "c59a89334cc15b9f89ace22ef30006af",
                "_score" : null,
                "_source" : {
                  "@timestamp" : "2021-03-31T14:23:08.307Z",
                  "doc" : {
                    "type" : "request_for_credentials",
                    "group" : "{{event_group_name}}",
                    "stack" : "events",
                    "captcha_is_valid" : true,
                    "sso_id" : "form",
                    "name" : "{{new_user_first_name}}",
                    "ip" : "172.19.0.1",
                    "position" : {
                      "lon" : "12.56738",
                      "lat" : "41.871939999999995"
                    },
                    "note" : "...",
                    "email" : "{{new_user_email}}",
                    "timestamp" : 1.617200588307E9
                  }
                },
                "sort" : [
                  1617200588307
                ]
              }
            ]
          }
        }
        """.replace('{{new_user_email}}', new_user_email) \
            .replace('{{new_user_first_name}}', new_user_first_name) \
            .replace('{{event_group_name}}', event_group_name) \
            .replace('{{event_password}}', event_password)

        es_response = json.loads(_es_response)

        mocked_url = ELASTICSEARCH_INDEX_URL
        self._set_mocked_response(mocked_url, es_response, method='POST')

        new_users_created = create_users_by_events()
        self.assertEqual(new_users_created, 1)

        new_users = User.objects.filter(email=new_user_email)
        new_user = new_users.first()
        self.assertEqual(new_user.email, new_user_email)
        self.assertEqual(new_user.sso_app_profile.first_name, new_user_first_name)
        self.assertEqual(new_users.count(), 1)

        email_addresses = EmailAddress.objects.filter(email=new_user_email)
        email_address = email_addresses.first()
        self.assertEqual(email_addresses.count(), 1)
        self.assertEqual(email_address.verified, email_verified)
        self.assertEqual(email_address.primary, True)

        # confirmation email
        if not email_verified:
            self.assertEqual(len(mail.outbox), 1)
            cofirmation_message = mail.outbox[0]
            recipients = cofirmation_message.recipients()
            self.assertIn(new_user_email, recipients, 'user email not in recipients')

        # check profile groups
        self.assertEqual(new_user.sso_app_profile.groups.filter(name=event_group_name).count(), 1)

        if default_password is not None:
            self.assertEqual(new_user.has_usable_password(), True)

        # check user can login
        client = self._get_client()
        response = client.post(
            reverse('account_login'),
            data=self._get_login_object(new_user_email, default_password)
        )
        self.assertEqual(response.status_code, status.HTTP_302_FOUND)
        self.assertEqual(response.url, '/')
        self.assertIsNotNone(self._get_response_jwt_cookie(response), 'response has no jwt cookie')

    @responses.activate
    def test_can_not_create_users_by_event_objects_if_exists(self):
        rfc_event_listener = RequestForCredentialsEventListener(confirm_email=False,
                                                                first_name_field='name')
        rfc_event_listener.save()

        new_user_email = 'pinco@example.com'
        new_user_first_name = 'Pinco'
        event_group_name = 'q-hack40-cps-bs'

        new_pass = self._get_random_pass()
        previous_user = self._get_new_user(password=new_pass, username=new_user_first_name, email=new_user_email)
        print('previous', previous_user)

        _es_response = """{
          "took" : 1,
          "timed_out" : false,
          "_shards" : {
            "total" : 1,
            "successful" : 1,
            "skipped" : 0,
            "failed" : 0
          },
          "hits" : {
            "total" : {
              "value" : 1,
              "relation" : "eq"
            },
            "max_score" : null,
            "hits" : [
              {
                "_index" : "events_private",
                "_type" : "_doc",
                "_id" : "c59a89334cc15b9f89ace22ef30006af",
                "_score" : null,
                "_source" : {
                  "@timestamp" : "2021-03-31T14:23:08.307Z",
                  "doc" : {
                    "type" : "request_for_credentials",
                    "group" : "{{event_group_name}}",
                    "stack" : "events",
                    "captcha_is_valid" : true,
                    "sso_id" : "form",
                    "name" : "{{new_user_first_name}}",
                    "ip" : "172.19.0.1",
                    "position" : {
                      "lon" : "12.56738",
                      "lat" : "41.871939999999995"
                    },
                    "note" : "...",
                    "email" : "{{new_user_email}}",
                    "timestamp" : 1.617200588307E9
                  }
                },
                "sort" : [
                  1617200588307
                ]
              }
            ]
          }
        }
        """.replace('{{new_user_email}}', new_user_email).replace('{{new_user_first_name}}', new_user_first_name).replace('{{event_group_name}}', event_group_name)
        es_response = json.loads(_es_response)
        mocked_url = ELASTICSEARCH_INDEX_URL
        self._set_mocked_response(mocked_url, es_response, method='POST')

        new_users_created = create_users_by_events()
        self.assertEqual(new_users_created, 0)

        users = User.objects.filter(email=new_user_email)
        self.assertEqual(users.count(), 1)



import logging

from django.urls import reverse
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.contrib.auth.hashers import get_hasher, identify_hasher

from ... import app_settings
from ..groups.models import Group

logger = logging.getLogger('django_sso_app.core.apps.users')


class DjangoSsoAppUserModelMixin(object):
    # username = models.CharField(_('username'), max_length=150, unique=True)  # django
    email = models.EmailField(_('email address'), blank=True)  # django

    """
    @property
    def previous_serialized_as_string(self):
        return getattr(self, '__serialized_as_string')

    @property
    def serialized_as_string(self):
        serialized = ''
        for f in app_settings.USER_FIELDS + ('password', 'is_active'):
            serialized += '__' + str(getattr(self, f, None))

        logger.debug('serialized_as_string {}'.format(serialized))
        return serialized
    """

    @property
    def password_has_been_updated(self):
        return getattr(self, '__dssoa__old_password', None) != self.password

    @property
    def email_has_been_updated(self):
        return getattr(self, '__dssoa__old_email', None) != self.email or \
               getattr(self, '__dssoa__email_updated', False)

    @property
    def sso_app_new_email(self):
        if self.email_has_been_updated:
            return self.email

    @property
    def username_has_been_updated(self):
        return getattr(self, '__dssoa__old_username', None) != self.username or \
               getattr(self, '__dssoa__username_updated', False)

    @property
    def sso_app_new_username(self):
        if self.username_has_been_updated:
            return self.username

    @property
    def password_has_been_hardened(self):
        has_password = getattr(self, '__dssoa__old_password', self.password) not in (None, '')

        try:
            if has_password:
                old_hashed_password = getattr(self, '__dssoa__old_password')
                old_hasher = identify_hasher(old_hashed_password)
                new_hashed_password = self.password
                new_hasher = identify_hasher(new_hashed_password)

                try:
                    hasher = get_hasher('default')
                    return hasher.must_update(old_hashed_password)

                except AssertionError:
                    logger.exception('Error checking password hardened')
                    return old_hasher != new_hasher
            else:
                return False

        except Exception:
            logger.exception('Exception on checking password hardened')
            return False

    def __init__(self, *args, **kwargs):
        super(DjangoSsoAppUserModelMixin, self).__init__(*args, **kwargs)
        if app_settings.BACKEND_ENABLED:
            setattr(self, '__dssoa__old_password', self.password)
            setattr(self, '__dssoa__old_email', self.email)
            setattr(self, '__dssoa__old_username', self.username)
            # setattr(self, '__serialized_as_string', self.serialized_as_string)  # replaced by '..__user_loggin_in'

    def save(self, *args, **kwargs):
        # pre save
        super(DjangoSsoAppUserModelMixin, self).save(*args, **kwargs)
        # post save
        #setattr(new_user, 'sso_app_profile', get_or_create_user_profile(new_user, commit=True))

    def get_relative_rest_url(self):
        profile = self.get_sso_app_profile()
        if profile is not None:
            return reverse("django_sso_app_user:rest-detail", kwargs={"sso_id": profile.sso_id})

    def get_sso_app_profile(self):
        return getattr(self, 'sso_app_profile', None)

    @property
    def sso_id(self):
        profile = self.get_sso_app_profile()
        if profile is not None:
            return profile.sso_id

        return None

    @property
    def sso_rev(self):
        profile = self.get_sso_app_profile()
        if profile is not None:
            return profile.sso_rev

        return 0

    @property
    def must_subscribe(self):
        profile = self.get_sso_app_profile()
        if profile is not None:
            return profile.must_subscribe

        return not self.is_active

    @property
    def is_subscribed(self):
        profile = self.get_sso_app_profile()
        if profile is not None:
            return profile.is_subscribed

        return self.is_active

    def add_to_group(self, group_name, creating=False):
        logger.info('Adding "{}" to group "{}"'.format(self, group_name))

        g, _created = Group.objects.get_or_create(name=group_name)
        if _created:
            logger.info('New group "{}" created'.format(g))

        if creating:
            setattr(self, '__dssoa__creating', True)

        if g not in self.groups.all():
            self.groups.add(g)

        if creating:
            setattr(self, '__dssoa__creating', False)

    def remove_from_group(self, group_name, creating=False):
        logger.info('Removing "{}" from group "{}"'.format(self, group_name))

        g, _created = Group.objects.get_or_create(name=group_name)
        if _created:
            logger.info('New group "{}" created'.format(g))

        if creating:
            setattr(self, '__dssoa__creating', True)

        if g in self.groups.all():
            self.groups.remove(g)

        if creating:
            setattr(self, '__dssoa__creating', False)



import logging

from django.contrib import auth

from ...permissions import is_authenticated
from ...utils import check_user_can_login, get_session_key
from ... import app_settings
from ...apps.groups.utils import set_default_user_groups, set_default_profile_groups

from .base import DjangoSsoAppAuthenticationBaseMiddleware

logger = logging.getLogger('django_sso_app.core.middleware')


class DjangoSsoAppAuthenticationBackendMiddleware(DjangoSsoAppAuthenticationBaseMiddleware):
    def backend_process_request(self, request, request_jwt, decoded_jwt):
        user = getattr(request, 'user', None)

        if not is_authenticated(request.user):
            # We are seeing this user for the first time in this session, attempt
            # to authenticate the user.
            if app_settings.APIGATEWAY_ENABLED:
                consumer_custom_id = get_session_key(request, '__dssoa__apigateway__consumer_custom_id')

                user = auth.authenticate(request=request,
                                         consumer_custom_id=consumer_custom_id,
                                         encoded_jwt=request_jwt,
                                         decoded_jwt=decoded_jwt)
            else:
                user = auth.authenticate(request=request,
                                         encoded_jwt=request_jwt,
                                         decoded_jwt=decoded_jwt)

            # set request.user
            setattr(request, 'user', user)  # !!

        # further checks
        if is_authenticated(user):
            check_user_can_login(user)

            if app_settings.MANAGE_USER_GROUPS:
                # default groups check
                set_default_user_groups(user)
                set_default_profile_groups(user.sso_app_profile)

            logger.info('User "{}" authenticated successfully!'.format(user))



import logging

import socket

from django.conf import settings

from rest_framework import status
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework.views import APIView


from ...core.permissions import is_staff
from ...core.apps.users.serializers import UserSerializer
from ...core.apps.groups.models import Group
from ...core.apps.groups.serializers import GroupSerializer
from ...core.apps.services.models import Service
from ...core.apps.services.serializers import ServiceSerializer

from ..views import local_space_available, CURRENT_DIR, logger

from django_sso_app import dist_name, __version__, app_settings

logger = logging.getLogger('django_sso_app.backend.api')


class StatsView(APIView):
    """
    Return instance stats
    """

    permission_classes = (AllowAny,)

    def get(self, request):
        try:
            health_status = 'green'

            user_types = Group.objects.filter(name__in=app_settings.USER_TYPES)
            services = Service.objects.filter(is_public=True)

            context = {
                'request': request
            }

            data = {
                'app_name': dist_name,
                'deployment_environment': settings.DEPLOYMENT_ENV,
                'revision': settings.REPOSITORY_REV,
                'version': __version__,

                'user': UserSerializer(request.user, context=context).data if request.user.is_authenticated else None,

                'user_types': GroupSerializer(user_types, many=True, context=context).data,

                'services': ServiceSerializer(services, many=True, context=context).data,

                'host': socket.gethostname()
            }

            if is_staff(request.user):
                free_space_mb = int(local_space_available(CURRENT_DIR) / (1024 * 1024))

                logger.info(
                    'Free space (MB): {}.'.format(free_space_mb))

                if free_space_mb > 200:
                    health_status = 'green'
                else:
                    if free_space_mb < 100:
                        health_status = 'yellow'
                    else:
                        health_status = 'red'

                data['free_space_mb'] = free_space_mb
                data['meta'] = str(request.META.items())

            data['status'] = health_status

            return Response(data, status=status.HTTP_200_OK)

        except Exception as e:
            err_msg = str(e)
            logger.exception('Error getting stats {}'.format(err_msg))
            return Response(err_msg, status.HTTP_500_INTERNAL_SERVER_ERROR)



from django.utils.deprecation import MiddlewareMixin

from .. import app_settings


class SameSiteMiddleware(MiddlewareMixin):

    def process_response(self, request, response):

        if app_settings.SAME_SITE_COOKIE_NONE:
            if 'jwt' in response.cookies:
                response.cookies['jwt']['samesite'] = 'None'

        return response



import logging

import os
import platform

from urllib.parse import urlsplit
from django.utils import translation
from django.shortcuts import redirect
# from django.views.generic import TemplateView

logger = logging.getLogger('django_sso_app.backend')

CURRENT_DIR = os.getcwd()


if platform.system() == 'Windows':
    def local_space_available(dir):
        """Return space available on local filesystem."""
        import ctypes
        free_bytes = ctypes.c_ulonglong(0)
        ctypes.windll.kernel32.GetDiskFreeSpaceExW(ctypes.c_wchar_p(dir), None, None, ctypes.pointer(free_bytes))
        return free_bytes.value
else:
    def local_space_available(dir):
        destination_stats = os.statvfs(dir)
        return destination_stats.f_bsize * destination_stats.f_bavail


def set_language_from_url(request, user_language):
    prev_lang = request.session.get(translation.LANGUAGE_SESSION_KEY, request.LANGUAGE_CODE)

    translation.activate(user_language)
    request.session[translation.LANGUAGE_SESSION_KEY] = user_language

    url = request.META.get('HTTP_REFERER', '/')
    parsed = urlsplit(url)

    _url = parsed.path
    if len(_url) > 1:
        _url = '/{}/'.format(user_language) + _url.lstrip('/{}/'.format(prev_lang))
    else:
        _url = '/'

    return redirect(_url)



import logging

from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.urls import reverse

from ..profiles.models import Profile
from ...models import CreatedAtModel, DeactivableModel

logger = logging.getLogger('django_sso_app.core.apps.devices')


class Device(CreatedAtModel, DeactivableModel):
    # deleting deletes jwt by delete_device_jwt signal!
    class Meta:
        app_label = 'django_sso_app'
        verbose_name = _('Device')

    profile = models.ForeignKey(
        Profile,
        on_delete=models.CASCADE,
        related_name="devices",
        verbose_name=_('profile')
    )

    fingerprint = models.CharField(_('fingerprint'), max_length=32)
    user_agent = models.CharField(_('user agent'), max_length=32, null=True, blank=True)

    apigw_jwt_id = models.CharField(_('apigw_jwt_id'), max_length=36, null=True, blank=True)
    apigw_jwt_key = models.CharField(_('apigw_jwt_key'), max_length=32, null=True, blank=True)
    apigw_jwt_secret = models.CharField(_('apigw_jwt_secret'), max_length=32, null=True, blank=True)

    def __str__(self):
        return '{}:{}:{}'.format(self.id, self.fingerprint, self.apigw_jwt_id)

    def get_relative_rest_url(self):
        return reverse("django_sso_app_device:rest-detail", args=[self.pk])

    def get_jwt_payload(self):
        return {
            'id': self.id,
            'fp': self.fingerprint,
            'sso_id': str(self.profile.sso_id),
            'sso_rev': self.profile.sso_rev,
            'iss': self.apigw_jwt_key,
            # 'iat': datetime.utcnow()
        }

    @property
    def user(self):
        return self.profile.user



import logging

from django.contrib.auth import get_user_model
from rest_framework import mixins
from rest_framework import permissions
from rest_framework import viewsets

from .models import Device
from .serializers import DeviceSerializer
from ...permissions import is_staff, is_authenticated, OwnerOrStaffPermission

User = get_user_model()

logger = logging.getLogger('django_sso_app.core.apps.devices')


class DeviceApiViewSet(mixins.ListModelMixin,
                       mixins.RetrieveModelMixin,
                       mixins.DestroyModelMixin,
                       viewsets.GenericViewSet):
    """
    Device API viewset
    """

    serializer_class = DeviceSerializer
    permission_classes = (OwnerOrStaffPermission,)

    def get_queryset(self):
        requesting_user = self.request.user

        if is_authenticated(requesting_user):
            if is_staff(requesting_user):
                return Device.objects.all()
            else:
                return Device.objects.filter(profile=requesting_user.sso_app_profile)

        return Device.objects.none()



# Boostable functions
#
# import pyximport
# pyximport.install()

from typing import Optional


def get_referer(APP_DOMAIN: str,
                 DJANGO_SSO_APP_BACKEND_DOMAINS_DICT: dict,
                 DJANGO_SSO_APP_BACKEND_DOMAINS: list,
                 DJANGO_SSO_APP_BACKEND_FULL_URLS_CHAIN: list) -> Optional[str]:
    """
    Extracts referer from request header
    :param referer:
    :param APP_DOMAIN
    :param DJANGO_SSO_APP_BACKEND_DOMAINS_DICT:
    :param DJANGO_SSO_APP_BACKEND_DOMAINS:
    :param DJANGO_SSO_APP_BACKEND_FULL_URLS_CHAIN:
    :return:
    """

    _index_previous_should_be = DJANGO_SSO_APP_BACKEND_DOMAINS_DICT[APP_DOMAIN] - 1
    if _index_previous_should_be < 0:
        _last_sso_instance = DJANGO_SSO_APP_BACKEND_DOMAINS[-1]
        if _last_sso_instance == APP_DOMAIN:
            referer = None
        else:
            referer = DJANGO_SSO_APP_BACKEND_FULL_URLS_CHAIN[-1]
            # Referrer was empty, setting as LAST SSO instance
    else:
        referer = DJANGO_SSO_APP_BACKEND_FULL_URLS_CHAIN[_index_previous_should_be]
        # Referrer was empty, setting as PREVIOUS SSO instance

    return referer


def get_next_bump(actual: str,
                  DJANGO_SSO_APP_BACKEND_FULL_URLS_CHAIN: list) -> Optional[str]:
    """
    Returns next django-sso-app backend instance
    :param actual:
    :param DJANGO_SSO_APP_BACKEND_FULL_URLS_CHAIN:
    :return:
    """
    if actual in DJANGO_SSO_APP_BACKEND_FULL_URLS_CHAIN:
        pos = DJANGO_SSO_APP_BACKEND_FULL_URLS_CHAIN.index(actual)

    last_pos = len(DJANGO_SSO_APP_BACKEND_FULL_URLS_CHAIN) - 1

    if pos < last_pos:
        next_pos = DJANGO_SSO_APP_BACKEND_FULL_URLS_CHAIN[pos + 1]
    else:
        next_pos = DJANGO_SSO_APP_BACKEND_FULL_URLS_CHAIN[0]

    if next_pos == actual:
        return None

    return next_pos



# Generated by Django 2.2.9 on 2020-04-07 21:54

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0002_email_max_length'),
        ('django_sso_app', '0002_auto_20200307_1035'),
    ]

    operations = [
        migrations.CreateModel(
            name='EmailAddress',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('account.emailaddress',),
        ),
    ]



import json

import jwt
from django.contrib.auth import get_user_model
from django.test import Client
from django.urls import reverse
from rest_framework import status

from django_sso_app.core.tests.factories import UserTestCase

User = get_user_model()


class TestApiGateway(UserTestCase):

    def check_apigateway_is_enabled(self):
        pass



import logging

import jwt

from ...tokens.utils import jwt_decode, jwt_encode
from ...utils import get_session_key, set_session_key, logger, set_cookie, get_random_fingerprint
from ...tokens.utils import get_request_jwt
from ...exceptions import RequestHasValidJwtWithNoDeviceAssociated
from ... import app_settings

logger = logging.getLogger('django_sso_app.core.apps.devices')


def add_profile_device(profile, fingerprint, secret=None):
    secret = secret or app_settings.TOKENS_JWT_SECRET

    logger.info(
        'Adding User Device for profile "{}" with fingerprint "{}" and secret "{}"'.format(profile,
                                                                                           fingerprint,
                                                                                           secret))

    device = profile.devices.model.objects.create(profile=profile, fingerprint=fingerprint, apigw_jwt_secret=secret)

    logger.debug('device "{}" created with key "{}" secret "{}" and fingerprint "{}"'.format(device,
                                                                                             device.apigw_jwt_key,
                                                                                             device.apigw_jwt_secret,
                                                                                             device.fingerprint))
    return device


def remove_profile_device(device):
    logger.info('Deleting Device "{}"'.format(device))

    device.delete()

    return 1


def remove_all_profile_devices(profile):
    logger.info('Removing All Profile Devices for "{}"'.format(profile))

    removed = 0
    for device in profile.devices.all():
        removed += remove_profile_device(device)

    return removed


def get_or_create_request_device(request):
    device = get_session_key(request, '__dssoa__device', None)

    if device is None:
        logger.debug('no previous device')
        fingerprint = get_session_key(request, '__dssoa__device__fingerprint')  # set by views

        if fingerprint is None:
            logger.debug('received empty fingerprint, checking JWT')

            raw_token = get_request_jwt(request)

            if raw_token is None:
                fingerprint = get_random_fingerprint(request)  # 'undefined'
            else:
                try:
                    device, verified_payload = jwt_decode(raw_token, verify=True)

                except RequestHasValidJwtWithNoDeviceAssociated:
                    logger.warning('no device associated to request token "{}"'.format(raw_token))
                    raise
                else:
                    fingerprint = verified_payload['fp']

        if device is None:
            user = request.user
            profile = user.sso_app_profile

            logger.info(
                'Request has no device, getting profile "{}" device with fingerprint "{}"'.format(profile,
                                                                                                  fingerprint))

            device = profile.devices.filter(fingerprint=fingerprint).first()

            if device is None:
                logger.debug('profile has no device with fingerprint "{}"'.format(fingerprint))
                device = add_profile_device(user.get_sso_app_profile(), fingerprint)
            else:
                logger.debug('fingerprint "{}" had device'.format(fingerprint))

        else:
            logger.debug('request has device "{}"'.format(device))

    else:
        logger.info(
            'request already has device "{}"'.format(device))

    assert device is not None

    set_session_key(request, '__dssoa__device', device)
    set_session_key(request, '__dssoa__device__fingerprint', device.fingerprint)

    return device


def renew_response_jwt(received_jwt, user, request, response):
    logger.debug('renewing response jwt for "{}"'.format(user))

    unverified_payload = jwt.decode(received_jwt, None,
                                    options={"verify_signature": False},
                                    algorithms=[app_settings.JWT_ALGORITHM])
    jwt_fingerprint = unverified_payload['fp']

    logger.info('Updating response JWT for User {0} with fingerprint {1}'.format(request.user, jwt_fingerprint))

    # creates new actual_device
    device = add_profile_device(user.get_sso_app_profile(), jwt_fingerprint)
    token = jwt_encode(device.get_jwt_payload(), device.apigw_jwt_secret)

    # max_age = 365 * 24 * 60 * 60  #one year
    # expires = datetime.datetime.strftime(datetime.datetime.utcnow() + \
    #     datetime.timedelta(seconds=max_age), "%a, %d-%b-%Y %H:%M:%S GMT")

    set_cookie(response, app_settings.JWT_COOKIE_NAME, token)



from django.contrib.auth import get_user_model

from django_sso_app.core.tests.factories import UserTestCase

from ...emails.models import EmailAddress
from ...groups.models import Group
from ...profiles.models import Profile

User = get_user_model()


class GroupTestCase(UserTestCase):
    def test_add_profile_to_group_updates_rev(self):
        user = self._get_new_user()

        group = Group.objects.create(name='new_group')

        profile = user.sso_app_profile

        profile_rev = profile.sso_rev

        profile.groups.add(group)

        print('profile groups', profile.groups.all())

        profile.refresh_from_db()

        self.assertEqual(profile.sso_rev, profile_rev + 1)

    def test_user_creation_adds_default_groups(self):
        group_name = self._get_random_string()
        default_user_group_names = [group_name]

        with self.settings(DJANGO_SSO_APP_DEFAULT_USER_GROUPS=default_user_group_names):
            user = self._get_new_user()

            user_groups = user.groups.values_list('name', flat=True)

            assert group_name in user_groups

    def test_user_profile_creation_adds_default_groups(self):
        group_name = self._get_random_string()
        default_profile_group_names = [group_name]

        with self.settings(DJANGO_SSO_APP_DEFAULT_PROFILE_GROUPS=default_profile_group_names):
            user = self._get_new_user()

            user_groups = user.sso_app_profile.groups.values_list('name', flat=True)

            print('GROP NAMES', user_groups, group_name)

            assert group_name in user_groups



from django_sso_app.core.tests.factories import UserTestCase

from django.contrib.auth import get_user_model
from django.urls import reverse
from django.test.utils import override_settings

from rest_framework import status


User = get_user_model()


class GroupTestCase(UserTestCase):
    @override_settings(DJANGO_SSO_APP_DEFAULT_USER_GROUPS=['group1', 'group2'])
    def test_user_groups_assigned_on_login(self):
        new_pass = self._get_random_pass()
        new_user = self._get_new_user(password=new_pass)

        new_group = self._get_new_group('group3')
        new_user.add_to_group(new_group)

        user_groups_before_login = new_user.groups.all()
        self.assertEqual(user_groups_before_login.count(), 3)

        client = self._get_client()
        response = client.post(
            reverse('account_login'),
            data=self._get_login_object(new_user.email, new_pass)
        )
        print('RES00', response)

        self.assertEqual(response.status_code, status.HTTP_302_FOUND)

        user_groups_after_login = new_user.groups.all()

        self.assertEqual(list(map(lambda x: x.name, user_groups_before_login)),
                         list(map(lambda x: x.name, user_groups_after_login)))

        response2 = client.post(
            reverse('account_logout')
        )

        self.assertEqual(self._get_response_jwt_cookie(response2), '', 'api logout keeps jwt cookie')




import logging

logger = logging.getLogger('django_sso_app.core.apps.groups')


def unconfirm_all_user_emails(user, request=None):
    logger.info('Unconfirming all emails for user "{}"'.format(user))

    return user.emailaddress_set.filter(verified=True).update(verified=False)



import pytest

from django_sso_app.backend.users.models import User

pytestmark = pytest.mark.django_db


def test_user_get_relative_url(user: User):
    assert user.get_relative_url() == f"/users/{user.username}/"



from django.views.generic import TemplateView

from .core import app_settings

urlpatterns = []
api_urlpatterns = []
i18n_urlpatterns = []


if app_settings.BACKEND_ENABLED:
    from .backend.urls import django_sso_app_i18n_urlpatterns, django_sso_app_urlpatterns, \
                              django_sso_app_api_urlpatterns
    urlpatterns += django_sso_app_urlpatterns
    api_urlpatterns += django_sso_app_api_urlpatterns
    i18n_urlpatterns += django_sso_app_i18n_urlpatterns

    if app_settings.BACKEND_STANDALONE:
        from django.contrib.staticfiles.urls import staticfiles_urlpatterns
        urlpatterns += staticfiles_urlpatterns()
else:
    from .app.urls import django_sso_app_api_urlpatterns, django_sso_app_urlpatterns, \
                          django_sso_app_i18n_urlpatterns
    urlpatterns += django_sso_app_urlpatterns
    api_urlpatterns += django_sso_app_api_urlpatterns
    i18n_urlpatterns += django_sso_app_i18n_urlpatterns

if app_settings.BACKEND_ENABLED or app_settings.APP_ENABLED:
    from .core.mixins import WebpackBuiltTemplateViewMixin

    assert(WebpackBuiltTemplateViewMixin is not None)
else:
    class WebpackBuiltTemplateView(TemplateView):
        pass



from django.urls import reverse

from rest_framework import serializers

from ...serializers import AbsoluteUrlSerializer, PartialObjectSerializer
from .models import Group, Grouping


class GroupSerializer(AbsoluteUrlSerializer):
    class Meta:
        model = Group
        read_only_fields = ('url', 'id', 'name')
        fields = read_only_fields


class GroupingListSerializer(AbsoluteUrlSerializer, PartialObjectSerializer):
    class Meta:
        model = Grouping
        read_only_fields = ('url', 'id', 'name', 'type', '_partial')
        fields = read_only_fields


class GroupingSerializer(AbsoluteUrlSerializer):
    group = serializers.SerializerMethodField(method_name='get_group_url')
    tree = serializers.SerializerMethodField(method_name='get_grouping_tree')

    class Meta:
        model = Grouping
        read_only_fields = ('url', 'id', 'name', 'type', 'group', 'description', 'tree')
        fields = read_only_fields

    def get_group_url(self, obj):
        request = self.context['request']
        _rel_rest_url = reverse("django_sso_app_group:rest-detail", args=[obj.name])

        return request.build_absolute_uri(_rel_rest_url)

    def get_grouping_tree(self, obj):
        def get_tree(n):
            node = {
                'node': {
                    'name': n.name,
                    'group': self.get_group_url(n),
                    'type': n.type,
                    'description': n.description,
                },
                'children': []
            }
            for c in n.get_children():
                node['children'].append(get_tree(c))

            return node

        return get_tree(obj)



import requests

from time import sleep
from jinja2 import Template


def get_es_query(template, **kwargs):
    return Template(template).render(**kwargs)


def call_backend(uri, es_query, jwt=None, apikey=None):
    """
    Send POST request to elasticsearch.
    """

    headers = {
        'Content-Type': 'application/json'
    }

    if jwt is not None:
        headers['Authorization'] = 'Bearer {}'.format(jwt)
    elif apikey is not None:
        headers['apikey'] = '{}'.format(apikey)

    try:
        resp = requests.post(uri, headers=headers, data=str(es_query))
        resp.raise_for_status()

    except Exception as e:
        # When status code is 502 elasticsearch is reindexing. Should retry.
        if e.response.status_code == 502:
            sleep(5)
            return call_backend(uri, es_query, jwt=jwt, apikey=apikey)
        else:
            raise

    else:
        return resp.json()



from django.conf import settings

from . import app_settings


def django_sso_app_context(request):
    return {
        "sso_cookie_domain": app_settings.JWT_COOKIE_NAME,

        "APP_DOMAIN": settings.APP_DOMAIN,
        "COOKIE_DOMAIN": app_settings.COOKIE_DOMAIN,
        "I18N_PATH_ENABLED": settings.I18N_PATH_ENABLED,

        "DJANGO_SSO_APP_BACKEND_ENABLED": app_settings.BACKEND_ENABLED,
        "DJANGO_SSO_APP_APIGATEWAY_ENABLED": app_settings.APIGATEWAY_ENABLED,
        "DJANGO_SSO_APP_JWT_COOKIE_NAME": app_settings.JWT_COOKIE_NAME,
        "DJANGO_SSO_APP_BACKEND_HAS_CUSTOM_FRONTEND_APP": app_settings.BACKEND_HAS_CUSTOM_FRONTEND_APP,
        "DJANGO_SSO_APP_PROFILE_FIELDS": app_settings.PROFILE_FIELDS,
        "DJANGO_SSO_APP_BACKEND_SIGNUP_MUST_FILL_PROFILE": app_settings.BACKEND_SIGNUP_MUST_FILL_PROFILE,
    }



from ...core.apps.users.forms import (UserCreationForm as DjangoSsoAppUserCreationForm,
                                      UserChangeForm as DjangoSsoAppUserChangeForm)


class UserChangeForm(DjangoSsoAppUserChangeForm):
    pass


class UserCreationForm(DjangoSsoAppUserCreationForm):
    pass



import logging

from django.core.exceptions import ObjectDoesNotExist
from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend

from ...apps.profiles.models import Profile

logger = logging.getLogger('django_sso_app.core.authentication.backends')
User = get_user_model()


class DjangoSsoAppApiGatewayBackendAuthenticationBackend(ModelBackend):
    """
    Authenticates by request CONSUMER_CUSTOM_ID header
    """

    backend_path = 'django_sso_app.core.authentication.backends.DjangoSsoAppApiGatewayAuthenticationBackend'

    def backend_authenticate(self, request, consumer_custom_id, encoded_jwt, decoded_jwt):
        logger.info('BACKEND authenticating by apigateway consumer {}'.format(consumer_custom_id))

        if consumer_custom_id is None:
            logger.debug('consumer_custom_id not set, skipping authentication')

            return

        try:
            sso_id = consumer_custom_id
            profile = Profile.objects.get(sso_id=sso_id)
            user = profile.user

        except ObjectDoesNotExist:
            logger.debug('user with apigateway consumer_custom_id "{}" does not exists'.format(consumer_custom_id))

            return

        else:
            logger.debug('user with apigateway consumer_custom_id "{}" exists'.format(consumer_custom_id))

            # allauth logic
            setattr(user, 'backend', self.backend_path)

        return user


class DjangoSsoAppJwtBackendAuthenticationBackend(ModelBackend):
    """
    Authenticates by request jwt
    """

    backend_path = 'django_sso_app.core.authentication.backends.DjangoSsoAppApiGatewayAuthenticationBackend'

    def backend_authenticate(self, request, encoded_jwt, decoded_jwt):
        logger.info('BACKEND authenticating by request jwt')

        try:
            sso_id = decoded_jwt['sso_id']
            profile = Profile.objects.get(sso_id=sso_id)
            user = profile.user

        except ObjectDoesNotExist:
            logger.debug('user with sso_id "{}" does not exists'.format(sso_id))
            return

        else:
            logger.debug('user with sso_id "{}" exists'.format(sso_id))

            # allauth logic
            setattr(user, 'backend', self.backend_path)

        return user



from typing import Any, Sequence

from django.contrib.auth import get_user_model
from factory import DjangoModelFactory, Faker, post_generation


class UserFactory(DjangoModelFactory):

    username = Faker("user_name")
    email = Faker("email")
    first_name = Faker("first_name")

    @post_generation
    def password(self, create: bool, extracted: Sequence[Any], **kwargs):
        password = (
            extracted
            if extracted
            else Faker(
                "password",
                length=42,
                special_chars=True,
                digits=True,
                upper_case=True,
                lower_case=True,
            ).generate(extra_kwargs={})
        )
        self.set_password(password)

    class Meta:
        model = get_user_model()
        django_get_or_create = ["username"]



import json
import logging

import requests

from django.db import transaction
from django.utils.encoding import smart_str
from django.contrib.auth import get_user_model

from ... import app_settings
from ..emails.models import EmailAddress
from ..profiles.utils import update_profile
from ..api_gateway.functions import get_apigateway_profile_groups_from_header

logger = logging.getLogger('django_sso_app.core.apps.users')


def fetch_remote_user(sso_id, encoded_jwt=None):
    """
    Fetches user model from remote django-sso-app backend
    :param sso_id:
    :param encoded_jwt:
    :return:
    """
    logger.info("Getting SSO profile for ID {} ...".format(sso_id))

    if encoded_jwt is None:
        logger.debug('Using Token')
        headers = {
            "Authorization": "Token {}".format(app_settings.BACKEND_STAFF_TOKEN)
        }
    else:
        logger.debug('Using JWT')
        headers = {
            "Authorization": "Bearer {}".format(encoded_jwt)
        }

    url = app_settings.REMOTE_USER_URL.format(sso_id=sso_id) + '?with_password=true'

    response = requests.get(url=url, headers=headers, timeout=10, verify=False)
    response.raise_for_status()
    sso_user = response.json()

    logger.info("Retrieved SSO profile for ID {}".format(sso_id))

    return sso_user


def update_user(user, update_object, commit=True):
    logger.info('Try update user fields for "{}"'.format(user))

    for f in app_settings.USER_FIELDS + ('password', 'is_active', 'date_joined'):
        if hasattr(user, f):
            new_val = update_object.get(f, None)
            if new_val is not None:
                setattr(user, f, new_val)
                logger.debug('user field updated "{}":"{}"'.format(f, new_val if f != 'password' else '*'))

    if commit:
        user.save()

    return user


def check_remote_user_existence_by_username(username):
    logger.info('Checking with SSO Backend if user {} already exists ...'
                .format(smart_str(username)))

    url = app_settings.BACKEND_USERS_CHECK_URL
    params = {
        'username': username
    }
    response = requests.get(url=url, params=params, timeout=10, verify=False)
    response.raise_for_status()

    if response.status_code == requests.codes.NOT_FOUND:
        logger.info('User {} does NOT exist'.format(smart_str(username)))

        return None

    sso_id = json.loads(response.text).get('sso_id', None)

    logger.info('User {} already exists with SSO ID {}'
                .format(smart_str(username), sso_id))

    return sso_id


def check_remote_user_existence_by_email(email):
    logger.info('Checking with SSO if a user with email {} already exists ...'
                .format(smart_str(email)))

    url = app_settings.REMOTE_USERS_CHECK_URL
    params = {
        'email': email
    }
    response = requests.get(url=url, params=params, timeout=10, verify=False)
    response.raise_for_status()

    if response.status_code == requests.codes.NOT_FOUND:
        logger.info('A user with email {} does NOT exist'
                    .format(smart_str(email)))

        return None

    sso_id = json.loads(response.text).get('sso_id', None)

    logger.info('A user with email {} already exists with SSO ID {}'
                .format(smart_str(email), sso_id))

    return sso_id


def get_remote_user_by_sso_id(sso_id):
    logger.info('Getting remote user with sso_id "{}"'.format(sso_id))

    url = app_settings.REMOTE_USER_URL.format(sso_id=sso_id) + '?with_password=true'
    params = {
        'sso_id': sso_id
    }
    response = requests.get(url=url, params=params, timeout=10, verify=False)
    response.raise_for_status()

    if response.status_code == requests.codes.NOT_FOUND:
        logger.info('A user with sso_id "{}" does NOT exist'.format(smart_str(sso_id)))
        return

    response_data = json.loads(response.text)
    sso_id = response_data.get('sso_id', None)

    if sso_id is None:
        logger.warning('No user with sso_id "{}" found'.format(sso_id))
        return

    logger.info('A user with sso_id "{}" exists'.format(smart_str(sso_id)))
    return response_data


@transaction.atomic
def create_local_user_from_object(user_object, verified=False):
    logger.info('Creating local user from object "{}". Verified: {}'.format(user_object, verified))

    username = user_object.get('username')
    sso_id = user_object.get('sso_id', None)
    sso_rev = user_object.get('sso_rev', 0)

    User = get_user_model()

    if sso_id is not None:
        sso_id = smart_str(sso_id)

        if User.objects.filter(sso_app_profile__sso_id=sso_id).count() > 0:
            raise Exception('Profile with sso_id "{}" already present'.format(sso_id))

        logger.info('Creating local user with sso_id "{}" and sso_rev "{}" ...'.format(sso_id, sso_rev))

    if User.objects.filter(username=username).count() > 0:
        raise Exception('User with username "{}" already present'.format(username))

    new_user = User()

    user_object_profile = user_object.get('profile', None)

    setattr(new_user, '__dssoa__creating', True)
    setattr(new_user, '__dssoa__profile__sso_id', sso_id)
    setattr(new_user, '__dssoa__profile__sso_rev', sso_rev)

    setattr(new_user, '__dssoa__profile__object', user_object_profile)

    # saving model there (commit) implies profile creation with pre-defined attrs sso_id, sso_rev
    new_user = update_user(new_user, user_object, commit=True)

    email = user_object.get('email', None)

    if email is not None:
        if EmailAddress.objects.filter(email__iexact=email).count() > 0:
            raise Exception('EmailAddress already present')

        created_email = EmailAddress.objects.create(user=new_user,
                                                    email=email,
                                                    verified=verified,
                                                    primary=True)
        logger.info(
            'A new EmailAddress for {0} has been created with id {1}'
            ''.format(
                email, created_email.id))

        if verified:
            logger.info(
                '{0} must NOT confirm email {1}'.format(new_user, email))
        else:
            logger.info(
                '{0} must confirm email {1}'.format(new_user, email))

    logger.info('local user created "{}"'.format(new_user))

    setattr(new_user, '__dssoa__creating', False)

    return new_user


@transaction.atomic
def create_local_user_from_apigateway_headers(request):
    consumer_custom_id = request.META.get(app_settings.APIGATEWAY_CONSUMER_CUSTOM_ID_HEADER)
    consumer_groups = request.META.get(app_settings.APIGATEWAY_CONSUMER_GROUPS_HEADER, None)

    logger.info('Creating local user from apigateway consumer username "{}"'.format(smart_str(consumer_custom_id)))

    sso_id = consumer_custom_id
    sso_rev = 0
    username = consumer_custom_id

    user_object = {
        'sso_id': sso_id,
        'sso_rev': sso_rev,
        'username': username
    }

    if consumer_groups is not None:
        user_object['profile'] = {
            'groups': get_apigateway_profile_groups_from_header(consumer_groups)
        }

    new_user = create_local_user_from_object(user_object)

    return new_user


@transaction.atomic
def create_local_user_from_jwt(decoded_jwt):
    logger.info('Creating local user from jwt')

    sso_id = decoded_jwt.get('sso_id')
    sso_rev = decoded_jwt.get('sso_rev')
    username = decoded_jwt.get('username', sso_id)

    user_object = {
        'sso_id': sso_id,
        'sso_rev': sso_rev,
        'username': username
    }

    new_user = create_local_user_from_object(user_object)

    #setattr(new_user, 'sso_app_profile', get_or_create_user_profile(new_user, Profile, profile_object))

    return new_user


@transaction.atomic
def create_local_user_from_remote_backend(remote_user_object, can_subscribe=False):
    # Creates new user only by profile object
    logger.info('Creating local user from remote backend')

    new_user = create_local_user_from_object(remote_user_object,
                                             verified=getattr(remote_user_object, 'email_verified', False))

    setattr(new_user, '__dssoa__remote_user', remote_user_object)  # noqa (subscription required)

    return new_user


@transaction.atomic
def update_local_user_from_remote_backend(user, remote_user_object, commit=True):
    logger.info('Updating local user with sso_id "{}" ...'.format(user.sso_id))

    user = update_user(user, remote_user_object, commit)

    setattr(user, '__dssoa__remote_user', remote_user_object)

    remote_object_profile = remote_user_object['profile']
    # update profile
    setattr(user, 'sso_app_profile', update_profile(user.sso_app_profile, remote_object_profile, commit))

    return user


@transaction.atomic
def replicate_remote_user(sso_id):
    logger.info('Replicating user with sso_id "{}" from remote backend'.format(sso_id))

    remote_user_object = fetch_remote_user(sso_id=sso_id)

    User = get_user_model()
    prev_user = User.objects.filter(sso_app_profile__sso_id=sso_id).first()

    if prev_user is None:
        user = create_local_user_from_remote_backend(remote_user_object)

    else:
        logger.info('User with sso_id "{}" already present, updating'.format(sso_id))
        user = update_local_user_from_remote_backend(prev_user, remote_user_object)

    return user



from django.contrib.auth import get_user_model
from django.urls import reverse
from django.core import mail

from rest_framework import status

from allauth.account.models import EmailConfirmationHMAC, EmailAddress

from django_sso_app.core.tests.factories import UserTestCase

from django_sso_app.core.apps.profiles.models import Profile

User = get_user_model()


class TestBackend(UserTestCase):

    def test_can_login_by_apigateway_header(self):
        with self.settings(DJANGO_SSO_APP_SHAPE='backend_only_apigateway'):
            new_pass = self._get_random_pass()
            new_user = self._get_new_user(password=new_pass)

            profile_url = reverse('django_sso_app_profile:rest-detail', args=(new_user.sso_id,))

            user_device = self._get_user_device(new_user)

            client = self._get_client()
            client.cookies = self._get_jwt_cookie(user_device)

            response = client.get(
                profile_url,
                content_type='application/json',
                HTTP_X_CONSUMER_CUSTOM_ID=new_user.sso_app_profile.sso_id
            )

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.data.get('sso_id'), new_user.sso_id, 'sso_id differs from header')
            self.assertEqual(Profile.objects.filter(sso_id=new_user.sso_id).count(), 1, 'no new user created')

    def test_redirect_to_profile_complete_if_profile_is_incomplete(self):
        BACKEND_URL = 'http://accounts.example.com'

        with self.settings(DJANGO_SSO_APP_SHAPE='backend_only',  APP_URL=BACKEND_URL):

            new_pass = self._get_random_pass()
            new_user = self._get_new_incomplete_user(password=new_pass)

            user_device = self._get_user_device(new_user)

            client = self._get_client()
            client.cookies = self._get_jwt_cookie(user_device)

            response = client.get(
                '/profile/',
                content_type='application/json',
                HTTP_X_CONSUMER_CUSTOM_ID=new_user.sso_app_profile.sso_id
            )

            self.assertEqual(response.status_code, status.HTTP_302_FOUND)
            self.assertEqual(response.url, '/profile/complete/')

    def test_redirect_to_profile_complete_if_profile_is_incomplete_apigateway(self):

        with self.settings(DJANGO_SSO_APP_SHAPE='backend_app_apigateway'):

            new_pass = self._get_random_pass()
            new_user = self._get_new_incomplete_user(password=new_pass)

            user_device = self._get_user_device(new_user)

            client = self._get_client()
            client.cookies = self._get_jwt_cookie(user_device)

            response = client.get(
                '/profile/',
                content_type='application/json',
                HTTP_X_CONSUMER_CUSTOM_ID=new_user.sso_app_profile.sso_id
            )

            self.assertEqual(response.status_code, status.HTTP_302_FOUND)
            self.assertEqual(response.url, '/profile/complete/')

    def test_email_verification_redirects_to_profile_completion(self):

        with self.settings(DJANGO_SSO_APP_SHAPE='backend_app'):
            signup_obj = self._get_signup_object()

            client = self._get_client()
            response = client.post(
                reverse('account_signup'),
                data=signup_obj
            )

            self.assertEqual(response.status_code, status.HTTP_302_FOUND)
            self.assertEqual(len(mail.outbox), 1)

            email_address = EmailAddress.objects.get(email=signup_obj['email'])
            email_confirmation = EmailConfirmationHMAC(email_address)

            response2 = client.get(reverse('account_confirm_email',
                                   args=[email_confirmation.key]),
                                   follow=True)

            self.assertEqual(response2.redirect_chain[-1][0], '/profile/complete/')

    def test_email_verification_redirects_to_profile_completion_apigateway(self):

        with self.settings(DJANGO_SSO_APP_SHAPE='backend_app'):
            signup_obj = self._get_signup_object()

            client = self._get_client()
            response = client.post(
                reverse('account_signup'),
                data=signup_obj,
            )

            self.assertEqual(response.status_code, status.HTTP_302_FOUND)
            self.assertEqual(len(mail.outbox), 1)

            email_address = EmailAddress.objects.get(email=signup_obj['email'])
            email_confirmation = EmailConfirmationHMAC(email_address)

            new_user = User.objects.get(email=signup_obj['email'])

            response2 = client.get(reverse('account_confirm_email', args=[email_confirmation.key]),
                                   HTTP_X_CONSUMER_CUSTOM_ID=new_user.sso_app_profile.sso_id,
                                   follow=True)

            self.assertEqual(response2.redirect_chain[-1][0], '/profile/complete/')



from django.contrib.auth import get_user_model

from django_filters import rest_framework as filters

User = get_user_model()


class UserFilter(filters.FilterSet):
    class Meta:
        model = User
        fields = {
            'username': ['exact'],
            'sso_app_profile__sso_id': ['exact'],
            'groups__name': ['exact'],
            'date_joined': ['exact', 'gt', 'lt', 'gte', 'lte'],
            'last_login': ['exact', 'gt', 'lt', 'gte', 'lte'],
        }



# Generated by Django 2.2.17 on 2021-04-01 16:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_sso_app', '0009_auto_20210319_1302'),
    ]

    operations = [
        migrations.CreateModel(
            name='RequestForCredentialsEventListener',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_active', models.BooleanField(default=True, verbose_name='is active')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
                ('updated_at', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='updated at')),
                ('meta', models.TextField(blank=True, null=True, verbose_name='Meta')),
                ('event_type', models.CharField(default='request_for_credentials', max_length=255)),
                ('confirm_email', models.BooleanField(default=False)),
                ('default_password', models.CharField(blank=True, default=None, max_length=255, null=True)),
                ('is_active_field', models.CharField(default='is_active', max_length=255)),
                ('username_field', models.CharField(default='username', max_length=255)),
                ('email_field', models.CharField(default='email', max_length=255)),
                ('password_field', models.CharField(default='password', max_length=255)),
                ('group_field', models.CharField(default='group', max_length=255)),
                ('sso_id_field', models.CharField(default='sso_id', max_length=255)),
                ('sso_rev_field', models.CharField(default='sso_rev', max_length=255)),
                ('external_id_field', models.CharField(default='external_id', max_length=255)),
                ('ssn_field', models.CharField(default='ssn', max_length=255)),
                ('phone_field', models.CharField(default='phone', max_length=255)),
                ('first_name_field', models.CharField(default='first_name', max_length=255)),
                ('last_name_field', models.CharField(default='last_name', max_length=255)),
                ('alias_field', models.CharField(default='alias', max_length=255)),
                ('description_field', models.CharField(default='description', max_length=255)),
                ('picture_field', models.CharField(default='picture', max_length=255)),
                ('birthdate_field', models.CharField(default='birthdate', max_length=255)),
                ('country_of_birth_field', models.CharField(default='country_of_birth', max_length=255)),
                ('latitude_field', models.CharField(default='position.lat', max_length=255)),
                ('longitude_field', models.CharField(default='position.lon', max_length=255)),
                ('country_field', models.CharField(default='country', max_length=255)),
                ('address_field', models.CharField(default='address', max_length=255)),
                ('language_field', models.CharField(default='language', max_length=255)),
                ('gender_field', models.CharField(default='gender', max_length=255)),
                ('company_name_field', models.CharField(default='company_name', max_length=255)),
                ('expiration_date_field', models.CharField(default='expiration_date', max_length=255)),
                ('device_fingerprint_field', models.CharField(default='device_fingerprint', max_length=255)),
                ('service_url_field', models.CharField(default='service_url', max_length=255)),
            ]
        ),
        migrations.AddField(
            model_name='profile',
            name='created_by_event',
            field=models.BooleanField(default=False, verbose_name='created by event'),
        ),
    ]



import logging

from rest_framework import serializers

from ...serializers import AbsoluteUrlSerializer
from .models import Device

logger = logging.getLogger('django_sso_app.core.apps.services')


class DeviceSerializer(AbsoluteUrlSerializer):
    profile = serializers.SerializerMethodField(required=False)

    class Meta:
        model = Device
        fields = ('url', 'id', 'is_active', 'created_at', 'profile', 'fingerprint')

    def get_profile(self, instance):
        request = self.context['request']
        profile_url = instance.profile.get_relative_rest_url()

        return request.build_absolute_uri(profile_url)



# -*- coding: utf-8 -*-
from pkg_resources import get_distribution, DistributionNotFound

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = 'django-sso-app'
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:
    __version__ = '0.11.12'
finally:
    del get_distribution, DistributionNotFound

# django
default_app_config = 'django_sso_app.apps.DjangoSsoAppConfig'



import logging

from django.forms import CharField, DateField, FloatField, HiddenInput

from allauth.account.forms import (LoginForm as allauth_LoginForm,
                                   SignupForm as allauth_SignupForm)
from allauth.account.adapter import get_adapter
from allauth.account import app_settings as allauth_app_settings

from .apps.emails.models import EmailAddress
from .apps.services.models import Service
from .apps.groups.models import Group

from .utils import get_random_fingerprint
from . import app_settings

logger = logging.getLogger('django_sso_app.core')


SIGNUP_FORM_FIELDS = app_settings.USER_FIELDS + (
    'email2',  # ignored when not present
    'password1',
    'password2'  # ignored when not present
)


class LoginForm(allauth_LoginForm):
    fingerprint = CharField(label='fingerprint', max_length=255, required=False, widget=HiddenInput())

    def __init__(self, *args, **kwargs):
        super(LoginForm, self).__init__(*args, **kwargs)

        request = kwargs.get('request', None)
        self.fields['fingerprint'].initial = get_random_fingerprint(request)

    def login(self, request, redirect_url=None):
        res = super(LoginForm, self).login(request, redirect_url)
        # check because of allauth author "misunderstanding" if user is trying to login with unconfirmed email:
        # https://github.com/pennersr/django-allauth/issues/254
        if allauth_app_settings.EMAIL_VERIFICATION == allauth_app_settings.EmailVerificationMethod.MANDATORY:
            login = self.cleaned_data['login']
            adapter = get_adapter(request)
            if self._is_login_email(login):
                emailaddress = EmailAddress.objects.filter(user=self.user,
                                                           email=login,
                                                           verified=False).first()
                if emailaddress is not None:
                    emailaddress.send_confirmation(request, signup=False)
                    res = adapter.respond_email_verification_sent(
                        request, self.user)

        return res


class BaseSignupForm(allauth_SignupForm):
    fingerprint = CharField(label='fingerprint', max_length=255, required=False, widget=HiddenInput())

    referer = CharField(label='referer', max_length=255, required=False, widget=HiddenInput(),
                        initial=app_settings.APP_URL)

    groups = CharField(label='groups', required=False)

    field_order = SIGNUP_FORM_FIELDS

    def __init__(self, *args, **kwargs):
        super(BaseSignupForm, self).__init__(*args, **kwargs)

        if 'username' not in app_settings.USER_FIELDS:
            self.fields.pop('username')
        if 'email' not in app_settings.USER_FIELDS:
            self.fields.pop('email')

        request = kwargs.get('request', None)
        self.fields['fingerprint'].initial = get_random_fingerprint(request)

    def get_form_kwargs(self):
        """Return the keyword arguments for instantiating the form."""
        kwargs = super(BaseSignupForm, self)
        return kwargs

    def save(self, request):
        new_user = super(BaseSignupForm, self).save(request)

        referer = self.cleaned_data["referer"]
        groups = self.cleaned_data.get("groups", '')

        try:
            if new_user is not None:
                # service subscription by "referer" field
                if referer not in (app_settings.APP_URL, '', None):
                    logger.info('Subscripting new user to service with url "{}"'.format(referer))

                    service = Service.objects.filter(service_url=referer).first()
                    if service is not None:
                        _subscription, _created = service.subscribe(new_user.get_sso_app_profile())

                # subscribing user to groups
                if len(groups):
                    group_names = groups.split(',')
                    logger.info('Subscripting new user to groups "{}"'.format(group_names))
                    for g in group_names:
                        group = Group.objects.filter(name=g).first()
                        if group is not None:
                            new_user.sso_app_profile.groups.add(group)


        # delete creating user model if any exception occurred
        except Exception as e:
            logger.exception('Error subscribing new user "{}"'.format(e))

            if new_user is not None:
                new_user.delete()

            raise

        return new_user


# custom signup form
if app_settings.BACKEND_SIGNUP_MUST_FILL_PROFILE:
    class SignupForm(BaseSignupForm):
        first_name = CharField(required='first_name' in app_settings.REQUIRED_PROFILE_FIELDS)
        last_name = CharField(required='last_name' in app_settings.REQUIRED_PROFILE_FIELDS)
        description = CharField(required='description' in app_settings.REQUIRED_PROFILE_FIELDS)
        picture = CharField(required='picture' in app_settings.REQUIRED_PROFILE_FIELDS)
        birthdate = DateField(required='birthdate' in app_settings.REQUIRED_PROFILE_FIELDS)
        latitude = FloatField(required='latitude' in app_settings.REQUIRED_PROFILE_FIELDS)
        longitude = FloatField(required='longitude' in app_settings.REQUIRED_PROFILE_FIELDS)
        country = CharField(required='country' in app_settings.REQUIRED_PROFILE_FIELDS)
        address = CharField(required='address' in app_settings.REQUIRED_PROFILE_FIELDS)
        # business_address = CharField(required='business_address' in app_settings.REQUIRED_PROFILE_FIELDS)
        language = CharField(required='language' in app_settings.REQUIRED_PROFILE_FIELDS)
        alias = CharField(required='alias' in app_settings.REQUIRED_PROFILE_FIELDS)

else:
    class SignupForm(BaseSignupForm):
        pass



import logging
import jwt

from rest_framework import HTTP_HEADER_ENCODING

from ..exceptions import RequestHasValidJwtWithNoDeviceAssociated
from .. import app_settings

_TOKEN_PREFIXES = tuple(map(lambda x: '{} '.format(x), app_settings.AUTH_HEADER_TYPES))
logger = logging.getLogger('django_sso_app.core.tokens')


def get_request_jwt_header(request, encoded=True):
    """
    Extracts the header containing the JSON web token from the given
    request.
    """
    jwt_header = request.META.get('HTTP_AUTHORIZATION', None)

    if encoded and isinstance(jwt_header, str):
        # Work around django test client oddness
        jwt_header = jwt_header.encode(HTTP_HEADER_ENCODING)

    logger.debug('header JWT "{}"'.format(jwt_header))
    return jwt_header


def get_request_jwt_cookie(request, encoded=True):
    """
    Extracts the header containing the JSON web token from the given
    request.
    """
    jwt_cookie = request.COOKIES.get(app_settings.JWT_COOKIE_NAME, None)

    if encoded and isinstance(jwt_cookie, str):
        # Work around django test client oddness
        jwt_cookie = jwt_cookie.encode(HTTP_HEADER_ENCODING)

    logger.debug('cookie JWT "{}"'.format(jwt_cookie))
    return jwt_cookie


def get_request_jwt(request, encoded=False):
    request_jwt = request_jwt_header = get_request_jwt_header(request, False)

    if request_jwt_header is None:
        request_jwt = get_request_jwt_cookie(request, False)

    else:
        _found = None
        for token_prefix in _TOKEN_PREFIXES:
            if request_jwt_header.find(token_prefix) != -1:
                _found = token_prefix
                break

        if _found is not None:
            logger.debug('request jwt found')
            request_jwt = request_jwt.replace(_found, '')
        else:
            logger.debug('"{}" absent on "{}"'.format(_TOKEN_PREFIXES, request_jwt_header))
            return

    if request_jwt is not None and encoded and isinstance(request_jwt, str):
        # Work around django test client oddness
        request_jwt = request_jwt.encode(HTTP_HEADER_ENCODING)

    return request_jwt


def get_request_jwt_fingerprint(request):
    received_jwt = get_request_jwt(request)
    if received_jwt is None:
        raise KeyError('No token specified')

    unverified_payload = jwt.decode(received_jwt, None,
                                    options={"verify_signature": False},
                                    algorithms=[app_settings.JWT_ALGORITHM])
    return unverified_payload.get('fp')


def jwt_decode_handler(token, secret_key, issuer, verify=True):
    # get user from token, BEFORE verification, to get user secret key

    unverified_payload = jwt.decode(token, None,
                                    options={"verify_signature": False},
                                    algorithms=[app_settings.JWT_ALGORITHM])

    if not verify:
        return unverified_payload

    options = {
        'verify_exp': False,
        'verify_signature': verify
    }

    return jwt.decode(
        token,
        secret_key,
        options=options,
        #leeway=api_settings.JWT_LEEWAY,
        #audience=api_settings.JWT_AUDIENCE,
        issuer=issuer,  # api_settings.APIJWT_ISSUER,
        algorithms=[app_settings.JWT_ALGORITHM]
    )


def jwt_encode_handler(payload, secret_key):
    return jwt.encode(
        payload,
        secret_key,
        app_settings.JWT_ALGORITHM
    )


def jwt_encode(payload, secret):
    return jwt_encode_handler(payload, secret)


def jwt_decode(raw_token, verify=True):
    if app_settings.BACKEND_ENABLED:
        from ..apps.devices.models import Device

        logger.debug('decode backend jwt')
        unverified_payload = jwt.decode(raw_token, None,
                                        options={"verify_signature": False},
                                        algorithms=[app_settings.JWT_ALGORITHM])

        try:
            device_id = unverified_payload['id']
            fingerprint = unverified_payload['fp']
            device = Device.objects.get(id=device_id,
                                        fingerprint=fingerprint)

        except Device.DoesNotExist:
            logger.warning('no device with "{}:{}" found in db, raising'.format(device_id, fingerprint))

            raise RequestHasValidJwtWithNoDeviceAssociated(device_id)

        return device, jwt_decode_handler(raw_token, device.apigw_jwt_secret, device.apigw_jwt_key, verify)

    else:
        logger.debug('decode app jwt')

        if app_settings.APIGATEWAY_ENABLED:
            return None, jwt_decode_handler(raw_token, app_settings.TOKENS_JWT_SECRET, None, False)
        else:
            return None, jwt_decode_handler(raw_token, app_settings.TOKENS_JWT_SECRET, None, verify)



from django.core.management.base import BaseCommand
from django.contrib.auth import get_user_model
from django.core.validators import validate_email
from django.core.exceptions import ValidationError

from django_sso_app.core.apps.users.utils import replicate_remote_user

User = get_user_model()


class Command(BaseCommand):
    help = 'Replicates a remote user'

    def add_arguments(self, parser):
        parser.add_argument('sso_id', type=str)

    def handle(self, *args, **options):
        sso_id = options['sso_id']

        new_user = replicate_remote_user(sso_id)

        self.stdout.write(self.style.SUCCESS('User "{}" replicated'.format(new_user)))



from django.db import models

from ...models import DeactivableModel, CreatedAtModel, MetaFieldModel, UpdatableModel


class EventListenerBaseModel(DeactivableModel, CreatedAtModel, UpdatableModel, MetaFieldModel):
    """
    EventListenerBaseModel

    """
    class Meta:
        abstract = True

    event_type = models.CharField(max_length=255, default='request_for_credentials')


class RequestForCredentialsEventListener(EventListenerBaseModel):
    """
    RequestForCredentialsEventListener

    """
    class Meta:
        app_label = 'django_sso_app'

    user_fields = ('username', 'email', 'password', 'is_active')
    profile_fields = ('sso_id', 'sso_rev', 'external_id', 'ssn', 'phone',
                      'first_name', 'last_name', 'alias', 'description', 'picture', 'birthdate', 'country_of_birth',
                      'latitude', 'longitude', 'country', 'address', 'language', 'gender', 'company_name',
                      'expiration_date')

    confirm_email = models.BooleanField(default=False)
    default_password = models.CharField(max_length=255, null=True, blank=True, default=None)

    # user
    is_active_field = models.CharField(max_length=255, default='is_active')

    username_field = models.CharField(max_length=255, default='username')
    email_field = models.CharField(max_length=255, default='email')
    password_field = models.CharField(max_length=255, default='password')

    # profile
    group_field = models.CharField(max_length=255, default='group')

    sso_id_field = models.CharField(max_length=255, default='sso_id')
    sso_rev_field = models.CharField(max_length=255, default='sso_rev')

    external_id_field = models.CharField(max_length=255, default='external_id')
    ssn_field = models.CharField(max_length=255, default='ssn')
    phone_field = models.CharField(max_length=255, default='phone')
    first_name_field = models.CharField(max_length=255, default='first_name')
    last_name_field = models.CharField(max_length=255, default='last_name')
    alias_field = models.CharField(max_length=255, default='alias')
    description_field = models.CharField(max_length=255, default='description')
    picture_field = models.CharField(max_length=255, default='picture')
    birthdate_field = models.CharField(max_length=255, default='birthdate')
    country_of_birth_field = models.CharField(max_length=255, default='country_of_birth')
    latitude_field = models.CharField(max_length=255, default='position.lat')
    longitude_field = models.CharField(max_length=255, default='position.lon')
    country_field = models.CharField(max_length=255, default='country')
    address_field = models.CharField(max_length=255, default='address')
    language_field = models.CharField(max_length=255, default='language')
    gender_field = models.CharField(max_length=255, default='gender')
    company_name_field = models.CharField(max_length=255, default='company_name')
    expiration_date_field = models.CharField(max_length=255, default='expiration_date')

    # device
    device_fingerprint_field = models.CharField(max_length=255, default='device_fingerprint')

    # service
    service_url_field = models.CharField(max_length=255, default='service_url')

    def __str__(self):
        return '{}:{}'.format(self.event_type, self.is_active)



import logging

from django.core.exceptions import ObjectDoesNotExist
from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend

from ...apps.users.utils import fetch_remote_user, create_local_user_from_remote_backend, \
                              update_local_user_from_remote_backend, create_local_user_from_jwt, \
                              create_local_user_from_apigateway_headers

from ...apps.profiles.models import Profile
from ... import app_settings


logger = logging.getLogger('django_sso_app.core.authentication.backends')
User = get_user_model()


class DjangoSsoAppAppBaseAuthenticationBackend(ModelBackend):

    def try_replicate_user(self, request, sso_id, encoded_jwt, decoded_jwt):
        logger.debug('try_replicate_user')

        if app_settings.REPLICATE_PROFILE:
            logger.info('Replicate user with sso_id "{}" from remote backend'.format(sso_id))

            # create local profile from SSO
            backend_user = fetch_remote_user(sso_id=sso_id, encoded_jwt=encoded_jwt)
            #backend_user_profile = backend_user['profile']

            user = create_local_user_from_remote_backend(backend_user)

            #if backend_user_profile.get('is_incomplete', False):
            #    redirect_to_profile_complete(user)

        else:
            user = None

        return user

    def try_update_user(self, sso_id, user, user_profile, encoded_jwt, decoded_jwt):
        logger.debug('try_update_user')

        rev_changed = user_profile.sso_rev < decoded_jwt['sso_rev']

        if rev_changed:
            if rev_changed:
                logger.info('Rev changed from "{}" to "{}" for user "{}", updating ...'
                            .format(user_profile.sso_rev, decoded_jwt['sso_rev'],
                                    user))

            # local profile updated from django_sso_app instance, do not update sso_rev
            setattr(user, '__dssoa__creating', True)

            remote_user_object = fetch_remote_user(sso_id=sso_id, encoded_jwt=encoded_jwt)
            user = update_local_user_from_remote_backend(user, remote_user_object)

            logger.info('{} updated with latest data from BACKEND'.format(user))

            setattr(user, '__dssoa__creating', False)

        else:
            logger.info('Nothing changed for user "{}"'.format(user))

        return user


class DjangoSsoAppApiGatewayAppAuthenticationBackend(DjangoSsoAppAppBaseAuthenticationBackend):
    """
    Authenticates by request CONSUMER_CUSTOM_ID header
    """

    def try_replicate_user(self, request, sso_id, encoded_jwt, decoded_jwt):
        user = super(DjangoSsoAppApiGatewayAppAuthenticationBackend, self).try_replicate_user(request, sso_id,
                                                                                              encoded_jwt, decoded_jwt)

        if user is None:
            logger.info('Creating user from headers')

            user = create_local_user_from_apigateway_headers(request)

        return user

    def app_authenticate(self, request, consumer_custom_id, encoded_jwt, decoded_jwt):
        logger.info('APP authenticating by apigateway consumer {}'.format(consumer_custom_id))

        try:
            sso_id = consumer_custom_id
            profile = Profile.objects.get(sso_id=sso_id)
            user = profile.user

        except ObjectDoesNotExist:
            logger.info('No profile with id "{}"'.format(sso_id))
            try:
                user = self.try_replicate_user(request, sso_id, encoded_jwt, decoded_jwt)

            except Exception as e:
                logger.exception('Can not replicate user: {}'.format(e))
                raise

        else:
            if app_settings.REPLICATE_PROFILE:
                logger.debug('Should replicate profile')

                if decoded_jwt is None:
                    logger.warning('decoded_jwt not set')
                    return

                user = self.try_update_user(sso_id, user, profile, encoded_jwt, decoded_jwt)

        return user


class DjangoSsoAppJwtAppAuthenticationBackend(DjangoSsoAppAppBaseAuthenticationBackend):
    """
    Authenticates by request jwt
    """

    def try_replicate_user(self, request, sso_id, encoded_jwt, decoded_jwt):

        user = super(DjangoSsoAppJwtAppAuthenticationBackend, self).try_replicate_user(request,
                                                                                       sso_id, encoded_jwt, decoded_jwt)

        if user is None:
            # create local profile from jwt
            logger.info('Replicating user with sso_id "{}" from JWT'.format(sso_id))

            user = create_local_user_from_jwt(decoded_jwt)

        return user

    def app_authenticate(self, request, encoded_jwt, decoded_jwt):
        logger.info('backend authenticating by request jwt')

        if encoded_jwt is None or decoded_jwt is None:
            logger.debug('request jwt not set, skipping authentication')

            return

        try:
            sso_id = decoded_jwt['sso_id']
            profile = Profile.objects.get(sso_id=sso_id)
            user = profile.user

            if app_settings.REPLICATE_PROFILE:
                logger.debug('try_update_user "{}" jwt consumer "{}"'.format(sso_id, sso_id))

                user = self.try_update_user(sso_id, user, profile, encoded_jwt, decoded_jwt)

            else:
                # just updates user groups
                logger.debug('Do not replicate profile')

        except ObjectDoesNotExist:
            try:
                user = self.try_replicate_user(request, sso_id, encoded_jwt, decoded_jwt)

            except Exception:
                logger.exception('Can not replicate remote user')
                raise

        else:
            if app_settings.REPLICATE_PROFILE:
                logger.debug('Should replicate profile')

                if decoded_jwt is None:
                    logger.warning('decoded_jwt not set')
                    raise

                user = self.try_update_user(sso_id, user, profile, encoded_jwt, decoded_jwt)

        return user



import logging

from django.db import models
from django.utils.translation import ugettext_lazy as _

from ...models import UuidPKModel, CreatedAtModel, UpdatableModel
from ... import app_settings

logger = logging.getLogger('django_sso_app.core.apps.status')


class Status(UuidPKModel, CreatedAtModel, UpdatableModel):
    """
    Should use django migrations to keep track of status switch
    """
    class Meta:
        app_label = 'django_sso_app'
        verbose_name = _('Status')

    shape = models.CharField(max_length=255, default=app_settings.SHAPE)

    def __str__(self):
        return self.shape



# register admin apps

import django_sso_app.core.apps.devices.admin
import django_sso_app.core.apps.groups.admin
import django_sso_app.core.apps.passepartout.admin
import django_sso_app.core.apps.profiles.admin
import django_sso_app.core.apps.services.admin
import django_sso_app.core.apps.users.admin
import django_sso_app.core.apps.events.admin



# Generated by Django 2.2.17 on 2021-02-22 15:26

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_sso_app', '0006_delete_emailaddress'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='passepartout',
            unique_together={('token', 'is_active')},
        ),
        migrations.RemoveField(
            model_name='passepartout',
            name='deleted',
        ),
    ]



from django.conf.urls import url
from django.conf.urls.static import static
from django.contrib import admin
from django.urls import include, path
from django.utils import timezone
from django.views import defaults as default_views
from django.views.decorators.http import last_modified
from django.contrib.flatpages.views import flatpage
from django.views.i18n import JavaScriptCatalog
from django.conf import settings
from django.conf.urls.i18n import i18n_patterns
from django.contrib.auth.decorators import login_required

from drf_spectacular.views import SpectacularAPIView, SpectacularSwaggerView

# sitemap
from django.contrib.sitemaps.views import sitemap
from django.contrib.flatpages.sitemaps import FlatPageSitemap

# django-sso-app
from ..urls import (urlpatterns as django_sso_app__urlpatterns,
                    api_urlpatterns as django_sso_app__api_urlpatterns,
                    i18n_urlpatterns as django_sso_app_i18n_urlpatterns)

from ..backend.views import set_language_from_url
from ..backend.api.views import StatsView
from ..backend.pages.views import HomeView
from .. import dist_name


last_modified_date = timezone.now()
js_info_dict = {}

urlpatterns = []
api_urlpatterns = []
_I18N_URLPATTERNS = []

urlpatterns += django_sso_app__urlpatterns
api_urlpatterns += django_sso_app__api_urlpatterns
_I18N_URLPATTERNS += django_sso_app_i18n_urlpatterns

urlpatterns += [
    # pai
    url(r'^i18n/', include('django.conf.urls.i18n')),
    url(r'^jsi18n/$', last_modified(lambda req, **kw: last_modified_date)(JavaScriptCatalog.as_view()), js_info_dict,
        name='javascript-catalog'),

    # Django Admin, use -% url 'admin:index' %-
    path(settings.ADMIN_URL, admin.site.urls),

    # set language url
    url(r'^set_language/(?P<user_language>\w+)/$', set_language_from_url, name="set_language_from_url"),

    # flatpages sitemap
    path('sitemap.xml', sitemap, {'sitemaps': {'flatpages': FlatPageSitemap()}},
         name='django.contrib.sitemaps.views.sitemap'),
]

_I18N_URLPATTERNS += [
    path('', login_required(HomeView.as_view(template_name='pages/home.html')), name='home'),

    # flatpages
    path('about/', flatpage, {'url': '/about/'}, name='about'),
    path('privacy-policy/', flatpage, {'url': '/privacy-policy/'}, name='privacy-policy'),
    path('legal-notes/', flatpage, {'url': '/legal-notes/'}, name='legal-notes'),
    path('media-policy/', flatpage, {'url': '/media-policy/'}, name='media-policy'),
]

if settings.I18N_PATH_ENABLED:
    urlpatterns += i18n_patterns(
        *_I18N_URLPATTERNS
    )
else:
    urlpatterns += _I18N_URLPATTERNS


# Your stuff: custom urls includes go here
urlpatterns += [
    path('users/', include('django_sso_app.backend.users.urls', namespace='users')),
]

urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)

if settings.DEBUG:
    # This allows the error pages to be debugged during development, just visit
    # these url in browser to see how these error pages look like.
    urlpatterns += [
        path(
            '400/',
            default_views.bad_request,
            kwargs={'exception': Exception('Bad Request!')},
        ),
        path(
            '403/',
            default_views.permission_denied,
            kwargs={'exception': Exception('Permission Denied')},
        ),
        path(
            '404/',
            default_views.page_not_found,
            kwargs={'exception': Exception('Page not Found')},
        ),
        path('500/', default_views.server_error),
    ]

    if 'debug_toolbar' in settings.INSTALLED_APPS:
        import debug_toolbar

        urlpatterns += [path('__debug__/', include(debug_toolbar.urls))]


api_urlpatterns += [
    url(r'^api/v1/_stats/$', StatsView.as_view(), name='stats'),

    # your api here
]

# extra
from .extra_urls import api_urlpatterns as extra_api_urlpatterns
if len(extra_api_urlpatterns):
    api_urlpatterns += extra_api_urlpatterns

urlpatterns += api_urlpatterns

urlpatterns += [
    path(
        "api/v1/",
        SpectacularSwaggerView.as_view(url_name="api-schema"),
        name="api-docs",
    ),
    path("api/v1/schema/", SpectacularAPIView.as_view(), name="api-schema"),
]

# extra

from .extra_urls import urlpatterns as extra_urlpatterns
from .extra_urls import i18n_urlpatterns as extra_i18n_urlpatterns

if len(extra_urlpatterns) > 0:
    urlpatterns += extra_urlpatterns

if settings.I18N_PATH_ENABLED:
    if len(extra_i18n_urlpatterns) > 0:
        urlpatterns += i18n_patterns(extra_i18n_urlpatterns)  #  + [])
else:
    if len(extra_i18n_urlpatterns) > 0:
        urlpatterns += extra_i18n_urlpatterns  # + []

# flatpages
urlpatterns += [
    path('<path:url>', include('django.contrib.flatpages.urls')),
]

# print('URLPATTERNS', urlpatterns)



import logging
import os

from django.utils.decorators import method_decorator
from django.views.decorators.debug import sensitive_post_parameters
from django.views.decorators.csrf import csrf_exempt
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from django.core.exceptions import SuspiciousOperation

from rest_framework import status, serializers
from rest_framework.views import APIView
from rest_framework.permissions import IsAuthenticated, AllowAny
from rest_framework.response import Response

from allauth.account.views import INTERNAL_RESET_SESSION_KEY
from allauth.account.forms import UserTokenForm
from allauth.utils import get_request_param
from allauth.account.adapter import get_adapter

from ..apps.emails.models import EmailAddress
from ..apps.users.serializers import SuccessfullLoginResponseSerializer, SuccessfullLogoutResponseSerializer
from ..apps.passepartout.utils import get_passepartout_login_redirect_url
from ..apps.emails.serializers import EmailSerializer
from ..permissions import is_authenticated
from ..utils import set_session_key, get_session_key, get_random_fingerprint
from .. import views as django_sso_app_views
from .. import app_settings

from .serializers import LoginSerializer, SignupSerializer, PasswordResetSerializer, PasswordResetFromKeySerializer, \
                         PasswordChangeSerializer, AddEmailSerializer, PasswordSetSerializer

sensitive_post_parameters_m = method_decorator(
    sensitive_post_parameters('password', 'password1', 'password2'))

logger = logging.getLogger('django_sso_app.core.api')

CURRENT_DIR = os.getcwd()
SUCCESS_STATUS_CODES = (status.HTTP_200_OK, status.HTTP_201_CREATED, status.HTTP_302_FOUND)
EMAIL_NOT_VERIFIED_MESSAGE = get_adapter().error_messages['email_not_verified']


class DjangoSsoAppApiViewMixin(APIView):
    # Form wrapper mixin
    http_method_names = ['get', 'post', 'head', 'options']

    __dssoa__is_api_view = True

    def __init__(self, *args, **kwargs):
        # object props
        self._initial = {}

        self.response_errors = 'Server error.'
        self.response_redirects_to_email_confirmation = False
        self.response_error_status = status.HTTP_500_INTERNAL_SERVER_ERROR

        super(DjangoSsoAppApiViewMixin, self).__init__(*args, **kwargs)

    def get_initial(self):
        """Return the initial data to use for forms on this view."""
        return self._initial.copy()

    def get_request_fingerprint(self, request):
        return request.data.get('fingerprint', get_random_fingerprint(request))

    def get_error_response(self):
        return Response(self.response_errors, status=self.response_error_status)

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        return super(DjangoSsoAppApiViewMixin, self).dispatch(request, *args, **kwargs)

    def form_valid(self, form):
        response = super(DjangoSsoAppApiViewMixin, self).form_valid(form)
        status_code = response.status_code

        if status_code not in SUCCESS_STATUS_CODES:
            raise Exception('Wrong status code "{}"'.format(status_code))

        location = response.get('location', None)
        if location == reverse('account_email_verification_sent'):
            self.response_redirects_to_email_confirmation = True

        return response

    def form_invalid(self, form):
        form_errors = dict(form.errors.items())

        if '__all__' in form_errors.keys():
            if len(form_errors['__all__']) == 1:
                parsed_errors = form_errors['__all__'][0]
            else:
                parsed_errors = form_errors['__all__']
        else:
            parsed_errors = form_errors

        self.response_error_status = status.HTTP_400_BAD_REQUEST
        self.response_errors = parsed_errors

        logger.debug('API Form errors "{}" -> "{}"'.format(form_errors, parsed_errors))

        raise serializers.ValidationError(parsed_errors)


# allauth views

class LoginView(DjangoSsoAppApiViewMixin, django_sso_app_views.LoginView):
    """
    Login
    """
    permission_classes = (AllowAny, )
    serializer_class = LoginSerializer
    http_method_names = ['post', 'head', 'options']
    email_not_verified_errors = EMAIL_NOT_VERIFIED_MESSAGE

    def get_form_kwargs(self, **kwargs):
        kwargs = super(LoginView, self).get_form_kwargs(**kwargs)

        kwargs['data'] ={
            'login': self.request.data.get('login', None),
            'password': self.request.data.get('password', None),
            'fingerprint': self.request.data.get('fingerprint', get_random_fingerprint(self.request)),
        }

        return kwargs

    def get_success_response(self):
        request = self.request
        data = {
            'user': request.user,
            'token': get_session_key(request, '__dssoa__jwt_token'),
            'redirect_url': get_passepartout_login_redirect_url(request),
        }
        serializer = SuccessfullLoginResponseSerializer(instance=data,
                                                        context={'request': self.request})

        response = Response(serializer.data, status=status.HTTP_200_OK)

        return response

    def post(self, request, *args, **kwargs):
        logger.info('API logging in')

        if is_authenticated(request.user):
            return Response(status=status.HTTP_400_BAD_REQUEST)

        try:
            set_session_key(request, '__dssoa__device__fingerprint', self.get_request_fingerprint(request))

            _response = super(LoginView, self).post(request, *args, **kwargs)
            _status_code = _response.status_code

            # check user must confirm email
            if self.response_redirects_to_email_confirmation:
                self.response_error_status = status.HTTP_400_BAD_REQUEST
                self.response_errors = self.email_not_verified_errors

                raise Exception('Email not verified')

        except Exception as e:
            logger.exception('Error: {}'.format(e))
            return self.get_error_response()

        else:
            if is_authenticated(request.user):
                response = self.get_success_response()
            else:
                logger.info('user "{}" is not logged in'.format(request.user))

                # check user has unsubscribed
                user_unsubscribed_at = get_session_key(request, '__dssoa__user_is_unsubscribed', None)
                if user_unsubscribed_at is not None:
                    self.response_error_status = status.HTTP_400_BAD_REQUEST
                    self.response_errors = 'Unsubscribed at "{}"'.format(user_unsubscribed_at)

                # check user must confirm email
                if _response.get('location', None) == reverse('account_email_verification_sent'):
                    self.response_error_status = status.HTTP_400_BAD_REQUEST
                    self.response_errors = self.email_not_verified_errors

                return self.get_error_response()

            return self.get_response_with_cookie(request, response)


class LogoutView(DjangoSsoAppApiViewMixin, django_sso_app_views.LogoutView):
    """
    Logout
    """
    permission_classes = (IsAuthenticated, )
    serializer_class = serializers.Serializer
    http_method_names = ['post', 'head', 'options']

    def get_success_response(self):
        request = self.request
        redirect_url = get_request_param(request, 'next')
        data = {
            'redirect_url': redirect_url
        }
        serializer = SuccessfullLogoutResponseSerializer(instance=data,
                                                         context={'request': self.request})

        return Response(serializer.data, status=status.HTTP_200_OK)

    def post(self, request, *args, **kwargs):
        logger.info('API Logging out')

        if request.user and request.user.is_anonymous:
            return Response(status=status.HTTP_400_BAD_REQUEST)

        try:
            _response = super(LogoutView, self).post(request, *args, **kwargs)
            _status_code = _response.status_code

            if _status_code not in SUCCESS_STATUS_CODES:
                logger.info('LogoutView status_code error ({})'.format(_status_code))

        except:
            return self.get_error_response()

        else:
            return self.get_response_with_invalidated_cookie(self.get_success_response())


class SignupView(DjangoSsoAppApiViewMixin, django_sso_app_views.SignupView):
    """
    Signup
    """
    permission_classes = (AllowAny, )
    serializer_class = SignupSerializer
    http_method_names = ['head', 'post', 'options']
    signup_form_fields = app_settings.USER_FIELDS + ('password1', 'password2',
                                                     'fingerprint',
                                                     'referer',
                                                     'groups')

    def get_form_kwargs(self, **kwargs):
        kwargs = super(SignupView, self).get_form_kwargs(**kwargs)

        data = {}
        for field in self.signup_form_fields:
            val = self.request.data.get(field, None)

            if field == 'groups':
                if val is not None and len(val):
                    val = ','.join(val)

            data[field] = val

        if app_settings.BACKEND_SIGNUP_MUST_FILL_PROFILE:
            for f in app_settings.PROFILE_FIELDS:
                data[f] = self.request.data.get(f, None)

        kwargs['data'] = data

        return kwargs

    def post(self, request, *args, **kwargs):
        logger.info('API Signing up')

        if is_authenticated(request.user):
            return Response(status=status.HTTP_400_BAD_REQUEST)

        try:
            _response = super(SignupView, self).post(request, *args, **kwargs)

        except:
            return self.get_error_response()

        else:
            return Response(_('We have sent an e-mail to you for verification. Follow the link provided to finalize the signup process. Please contact us if you do not receive it within a few minutes.'), status=status.HTTP_200_OK)


class EmailView(DjangoSsoAppApiViewMixin, django_sso_app_views.EmailView):
    """
    Manage user email addresses
    """
    permission_classes = (IsAuthenticated, )
    serializer_class = AddEmailSerializer
    http_method_names = ['get', 'post', 'head', 'options']

    def get_form_kwargs(self, **kwargs):
        kwargs = super(EmailView, self).get_form_kwargs(**kwargs)

        kwargs['data'] = {
            'email': self.request.data.get('email', None)
        }

        return kwargs

    def get(self, request, *args, **kwargs):
        logger.info('API Getting email')

        try:
            _response = super(EmailView, self).get(request, *args, **kwargs)

        except:
            return self.get_error_response()

        else:
            email_addresseses = EmailAddress.objects.filter(user=request.user)

            serializer = EmailSerializer(email_addresseses, many=True)

            return Response(serializer.data, status=status.HTTP_200_OK)

    def post(self, request, *args, **kwargs):
        logger.info('API adding new email')

        # set action_add as default
        request._request.POST = request._request.POST.copy()
        request._request.POST["action_add"] = ''

        try:
            _response = super(EmailView, self).post(request._request, *args, **kwargs)

        except:
            return self.get_error_response()

        else:
            return Response(_('We have sent an e-mail to you for verification. Please click on the link inside this e-mail. Please contact us if you do not receive it within a few minutes.'),
                            status=status.HTTP_200_OK)


class PasswordResetView(DjangoSsoAppApiViewMixin, django_sso_app_views.PasswordResetView):
    """
    Reset user password
    """
    permission_classes = (AllowAny, )
    serializer_class = PasswordResetSerializer
    http_method_names = ['head', 'post', 'options']

    def get_form_kwargs(self, **kwargs):
        kwargs = super(PasswordResetView, self).get_form_kwargs(**kwargs)

        kwargs['data'] = {
            'email': self.request.data.get('email', None)
        }

        return kwargs

    def post(self, request, *args, **kwargs):
        logger.info('API Asking for password reset')
        try:
            _response = super(PasswordResetView, self).post(request, *args, **kwargs)

        except:
            return self.get_error_response()

        else:
            return Response(_('We have sent you an e-mail. Please contact us if you do not receive it within a few minutes.'),
                            status=status.HTTP_200_OK)


class PasswordResetFromKeyView(DjangoSsoAppApiViewMixin, django_sso_app_views.PasswordResetFromKeyView):
    """
    Confirm user password reset
    """
    permission_classes = (AllowAny, )
    serializer_class = PasswordResetFromKeySerializer
    http_method_names = ['head', 'post', 'options']

    def get_form_kwargs(self, **kwargs):
        self.key = self.request.data.get('key')
        uidb36 = self.request.data.get('uidb36')

        token_form = UserTokenForm(
            data={'uidb36': uidb36, 'key': self.key})

        if token_form.is_valid():
            self.reset_user = token_form.reset_user

        else:
            logger.error('invalid password reset key')

            self.response_error_status = status.HTTP_400_BAD_REQUEST
            self.response_errors = _('The password reset link was invalid, possibly because it has already been used.  Please request a <a href="{}">new password reset</a>.'.format(reverse('account_reset_password')))

            raise SuspiciousOperation('Invalid password reset tokens')

        kwargs = super(PasswordResetFromKeyView, self).get_form_kwargs(**kwargs)

        kwargs['data'] = {
            'password1': self.request.data.get('password1', None),
            'password2': self.request.data.get('password2', None)
        }

        return kwargs

    def post(self, request, **kwargs):
        logger.info('Resetting password form key')
        try:
            key = self.request.data.get('key')
            uidb36 = self.request.data.get('uidb36')

            set_session_key(request, INTERNAL_RESET_SESSION_KEY, key)

            _response = super(PasswordResetFromKeyView, self).post(request, uidb36, key, **kwargs)

        except:
            logger.exception('Error while resetting password form key')
            return self.get_error_response()

        else:
            return Response(_('Your password is now changed.'), status=status.HTTP_200_OK)


class PasswordChangeView(DjangoSsoAppApiViewMixin, django_sso_app_views.PasswordChangeView):
    """
    Change user password
    """
    permission_classes = (IsAuthenticated, )
    serializer_class = PasswordChangeSerializer
    http_method_names = ['head', 'post', 'options']

    def get_form_kwargs(self, **kwargs):
        kwargs = super(PasswordChangeView, self).get_form_kwargs(**kwargs)

        kwargs['data'] = {
            'oldpassword': self.request.data.get('oldpassword', None),
            'password1': self.request.data.get('password1', None),
            'password2': self.request.data.get('password2', None)
        }

        return kwargs

    def post(self, request, *args, **kwargs):
        logger.info('API Asking for password change')

        try:
            _response = super(PasswordChangeView, self).post(request, *args, **kwargs)

        except:
            return self.get_error_response()

        else:
            return Response(_('Password updated, <a href=\'/login/\'>LOGIN</a> with new credentials.'),
                            status=status.HTTP_200_OK)


class PasswordSetView(DjangoSsoAppApiViewMixin, django_sso_app_views.PasswordSetView):
    """
    Set user password
    """
    permission_classes = (IsAuthenticated, )
    serializer_class = PasswordSetSerializer
    http_method_names = ['head', 'post', 'options']

    def get_form_kwargs(self, **kwargs):
        kwargs = super(PasswordSetView, self).get_form_kwargs(**kwargs)

        kwargs['data'] = {
            'password1': self.request.data.get('password1', None),
            'password2': self.request.data.get('password2', None)
        }

        return kwargs

    def post(self, request, *args, **kwargs):
        logger.info('API Asking for password set')

        if self.request.user.has_usable_password():
            return Response(_('Password already set.'), status=status.HTTP_400_BAD_REQUEST)

        try:
            _response = super(PasswordSetView, self).post(request, *args, **kwargs)

        except:
            return self.get_error_response()

        else:
            response = Response(_('Password successfully set.'),
                            status=status.HTTP_200_OK)

            self.update_response_jwt(request, response)

            return response



import logging

from django.contrib.auth import get_user_model
from django.shortcuts import redirect
from django.urls import reverse, reverse_lazy
from django.views.generic import UpdateView
from django.contrib.auth.mixins import LoginRequiredMixin

from rest_framework import permissions
from rest_framework import viewsets
from rest_framework import generics

from django_filters.rest_framework import DjangoFilterBackend

from .forms import UserProfileForm
from .models import Profile
from .serializers import ProfileSerializer
from ... import app_settings
from ...permissions import is_staff
from ...mixins import WebpackBuiltTemplateViewMixin
from ...views import DjangoSsoAppBaseViewMixin

from .filters import ProfileFilter

User = get_user_model()

logger = logging.getLogger('django_sso_app.core.apps.profiles')


class ProfileView(LoginRequiredMixin, WebpackBuiltTemplateViewMixin):
    template_name = "account/profile.html"

    def get_object(self, queryset=None):
        request_user = getattr(self.request, 'user', None)

        return request_user.get_sso_app_profile()

    def get_context_data(self, *args, **kwargs):
        context = super(ProfileView, self).get_context_data(*args, **kwargs)
        profile = self.get_object()

        if profile is not None:
            user_fields = []
            for field in app_settings.PROFILE_FIELDS:
                user_fields.append((field, getattr(profile, field)))

            context['user_fields'] = user_fields

        return context

    def get(self, request, *args, **kwargs):
        logger.info('Getting profile')

        try:
            return super(ProfileView, self).get(request, *args, **kwargs)

        except:
            logger.exception('Error getting profile')
            raise


class ProfileUpdateView(LoginRequiredMixin, UpdateView, WebpackBuiltTemplateViewMixin,
                        DjangoSsoAppBaseViewMixin):
    model = Profile
    template_name = "account/profile_update.html"
    form_class = UserProfileForm

    success_url = reverse_lazy('profile')

    def get_object(self, queryset=None):
        request_user = getattr(self.request, 'user', None)

        return request_user.get_sso_app_profile()

    def get_context_data(self, *args, **kwargs):
        self.object = self.get_object()

        context = super(WebpackBuiltTemplateViewMixin, self).get_context_data(*args, **kwargs)
        return context

    def get(self, request, *args, **kwargs):
        logger.info('Getting profile update')

        try:
            profile = self.get_object()
            return super(ProfileUpdateView, self).get(request, profile.pk, *args, **kwargs)

        except:
            logger.exception('Error getting profile update')
            raise

    def post(self, request, *args, **kwargs):
        logger.info('Posting profile update')

        try:
            profile = self.get_object()
            response = super(ProfileUpdateView, self).post(request, profile.pk, *args, **kwargs)

        except:
            logger.exception('Error posting profile update')
            raise

        else:
            user = profile.user
            self.update_response_jwt(user, response)

            return response


class ProfileCompleteView(ProfileUpdateView):
    template_name = "account/profile_complete.html"

    def get(self, request, *args, **kwargs):
        logger.info('Getting profile complete')

        if not request.user.sso_app_profile.is_incomplete:
            return redirect(reverse('profile.update'))

        return super(ProfileCompleteView, self).get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        logger.info('Posting profile create')

        if not request.user.sso_app_profile.is_incomplete:
            return redirect(reverse('profile.update'))

        return super(ProfileCompleteView, self).post(request, *args, **kwargs)


# api

class ProfileApiViewSet(viewsets.ModelViewSet,
                        DjangoSsoAppBaseViewMixin):
    lookup_field = 'sso_id'
    permission_classes = (permissions.IsAuthenticated, )
    filter_backends = (DjangoFilterBackend,)
    filterset_class = ProfileFilter
    serializer_class = ProfileSerializer

    def get_queryset(self):
        requesting_user = self.request.user
        queryset = Profile.objects.filter(user__is_superuser=False, is_active=True, user__is_active=True)

        if is_staff(requesting_user):
            return queryset
        else:
            return queryset.filter(user=requesting_user)

    def list(self, request, *args, **kwargs):
        """
        List profiles
        """
        return super(ProfileApiViewSet, self).list(request, *args, **kwargs)

    def retrieve(self, request, *args, **kwargs):
        """
        Profile detail
        """
        return super(ProfileApiViewSet, self).retrieve(request, *args, **kwargs)

    def partial_update(self, request, *args, **kwargs):
        """
        Profile update
        """
        response = super(ProfileApiViewSet, self).partial_update(request, *args, **kwargs)
        response_status_code = response.status_code
        logger.info('Profile partial_update status code is {}, {}'.format(response_status_code, request.data))

        if response_status_code == 200:
            profile = self.get_object()
            user = profile.user

            self.update_response_jwt(user, response)

        return response


class ProfileDetailApiView(generics.RetrieveAPIView):
    """
    Profile detail entrypoint
    """

    lookup_field = 'sso_id'
    serializer_class = ProfileSerializer

    def get_queryset(self):
        return Profile.objects.filter(is_active=True, user__is_active=True)

    def get_object(self, *args, **kwargs):
        user = getattr(self.request, 'user', None)

        return user.get_sso_app_profile()



import os
import json
import shutil

from django.urls import reverse
from django.contrib.auth import get_user_model
from django.conf import settings
from django.contrib import auth

from rest_framework import status
from rest_framework.test import APITestCase

from ..apps.emails.models import EmailAddress
from .settings import USERS_FOLDER

User = get_user_model()


class APITestFactory(APITestCase):
    def setUp(self):
        self.user_username = 'pippo'
        self.user_email = 'pippo@example.com'
        self.user_password = 'password'
        self.user = User.objects.create(username=self.user_username, email=self.user_email)
        self.user.set_password(self.user_password)
        self.user.save()

        self.user_email = EmailAddress.objects.create(user=self.user,
                                                      email=self.user_email,
                                                      primary=True,
                                                      verified=True)

        self.user_folder_path = os.path.join(settings.PRIVATE_ROOT, USERS_FOLDER, str(self.user.sso_id))

        self.valid_login = {
            'login': self.user_username,
            'password': self.user_password,
            'fingerprint': '123456'
        }

    def tearDown(self):
        shutil.rmtree(self.user_folder_path, ignore_errors=True)

    def perform_user_login(self):
        response = self.client.post(
            reverse('rest_login'),
            data=json.dumps(self.valid_login),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        user = auth.get_user(self.client)
        self.assertTrue(user.is_authenticated)

        return response



import logging

from urllib.parse import urlencode

# import pyximport
# pyximport.install()

from django.urls import reverse

from ...utils import get_session_key, set_session_key
from ...permissions import is_django_staff
from ... import app_settings
from ..devices.utils import get_or_create_request_device

from .models import Passepartout
from .functions import get_next_bump

logger = logging.getLogger('django_sso_app.core.apps.passepartout')


def get_passepartout_login_redirect_url(request):
    if is_django_staff(request.user):
        return None

    redirect_url = None

    if app_settings.PASSEPARTOUT_PROCESS_ENABLED:
        redirect_url = get_session_key(request, '__dssoa__passepartout__redirect_url')

        if redirect_url is None:
            logger.info(
                'Passepartout login enabled, BACKEND_URLS_CHAIN: {0}'.format(
                    app_settings.BACKEND_URLS_CHAIN))

            _nextUrl = nextUrl = request.GET.get('next', request.POST.get('next', None))

            if nextUrl is None:
                nextUrl = app_settings.APP_URL

            device = get_or_create_request_device(request)

            assert device is not None

            logger.info(
                'Login started passepartout logic '
                'for user {0}, with device {1}, nextUrl is {2}'.format(
                    request.user, device, nextUrl))

            token = get_session_key(request, '__dssoa__jwt_token')

            passepartout = Passepartout.objects.create_passepartout(device=device, jwt=token)
            logger.info('Created passepartout object {0}'.format(passepartout))

            args = {
                'next': nextUrl
            }

            next_sso_service = get_next_bump(app_settings.APP_URL, app_settings.BACKEND_FULL_URLS_CHAIN)
            if next_sso_service is not None:
                redirect_url = next_sso_service + reverse('django_sso_app_passepartout:login',
                                                          args=[passepartout.token]) + '?' + urlencode(args)

                set_session_key(request, '__dssoa__passepartout__redirect_url', redirect_url)

    else:
        logger.info('Only one SSO instance, no passepartout process')

    return redirect_url
